#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"
#include "texture.h"

#define ROTXMAX (PIHALF - 0.1f)

// These two external function pointers are exported by the VU code
// and we can now use them as pointers to the compiled VU data
extern "C" void VU_vu1code_start(void);
extern "C" void VU_vu1code_end(void);

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

void AddCube()
{
	// The first face draws two triangle
	// Next faces draw four triangles
	// Some triangles are "zero" area and will be culled by GS
	int iTriangles = 2 + (4 * 3);
	int iVerts = iTriangles + 2;

	// We want to unpack 3 quad words of misc data, and then 3QW per vertex starting at
	// VU mem location 8 (we will upload the matrices to the first 8QW per frame).
	VIFStaticDMA.AddUnpack(V4_32, 8, 9 + iVerts * 3);
	VIFStaticDMA.AddVector(Vector4(2048.0f, 2048.0f, ((float)0xFFFFFF) / 32.0f, *((float *)&iVerts))); //8

	// Set up 3 directional lights
	Vector4 LightDir1( 1,  0,  0,  0);	
	Vector4 LightDir2(-1,  0,  0,  0);
	Vector4 LightDir3( 0,  0, -1,  0);

	// Then normalise.
	LightDir1.NormaliseSelf();
	LightDir2.NormaliseSelf();
	LightDir3.NormaliseSelf();

	// Light dir matrix (on VU1 we can do 3 lights + Ambient for the same price as 1)
	VIFStaticDMA.AddVector(Vector4(LightDir1.x,LightDir2.x,LightDir3.x,0));				// 9
	VIFStaticDMA.AddVector(Vector4(LightDir1.y,LightDir2.y,LightDir3.y,0));				// 10
	VIFStaticDMA.AddVector(Vector4(LightDir1.z,LightDir2.z,LightDir3.z,0));				// 11

	// Light colour matrix (3 directional light colours)
	VIFStaticDMA.AddVector(Vector4(128,  0,  0,  0));									// 12
	VIFStaticDMA.AddVector(Vector4(0  ,128,  0,  0));									// 13
	VIFStaticDMA.AddVector(Vector4(0  ,  0,128,  0));									// 14
	// Ambient light colour
	VIFStaticDMA.AddVector(Vector4( 10, 10, 10,  0));									// 15


	// Add the GIFTag
	VIFStaticDMA.Add128(GS_GIFTAG_BATCH(	iVerts,						// NLOOP		// 16
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_TRI_STRIP, 	// PRIM
													PRIM_IIP_GOURAUD, 
													PRIM_TME_ON,
													PRIM_FGE_OFF, 
													PRIM_ABE_ON, 
													PRIM_AA1_OFF, 
													PRIM_FST_STQ, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_3(	GIF_REG_ST,
	 													GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_XYZ2)));
	
	
	
	// Front face
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 0, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1, 1, 1));	// Vert

	// Right face
	VIFStaticDMA.AddVector(Vector4( 1, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 1, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1,-1, 1));	// Vert

	// Back face
	VIFStaticDMA.AddVector(Vector4( 2, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 2, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 1,-1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1,-1, 1));	// Vert

	// Left face
	VIFStaticDMA.AddVector(Vector4( 3, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 3, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1,-1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 4, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1, 1, 1, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 4, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-1,-1, 1, 1));	// Vert
	
	
	// Flush to make sure VU1 isn't doing anything.
	VIFStaticDMA.Add32(FLUSH);
	// Then run the microcode.
	VIFStaticDMA.Add32(MSCALL(0));

	// Return back to the dynamic chain (note that we aren't aligned on a QWORD boundary, but
	// it doesn't matter because the packet class will sort that out for us).
	VIFStaticDMA.DMARet();
}

void AddAxes()
{	
	// 3 lines with two vertices for each line
	int iLines = 3;
	int iVerts = iLines * 2;

	// We want to unpack 9 quad words of misc data, and then 3QW per vertex starting at
	// VU mem location 8 (we will upload the matrices to the first 8QW per frame).
	VIFStaticDMA.AddUnpack(V4_32, 8, 9 + iVerts * 3);
	VIFStaticDMA.AddVector(Vector4(2048.0f, 2048.0f, ((float)0xFFFFFF) / 32.0f, *((float *)&iVerts))); //8

	// Set up 3 directional lights
	Vector4 LightDir1( 1,  0,  0,  0);	
	Vector4 LightDir2( 0,  1,  0,  0);
	Vector4 LightDir3( 0,  0,  1,  0);

	// Then normalise.
	LightDir1.NormaliseSelf();
	LightDir2.NormaliseSelf();
	LightDir3.NormaliseSelf();

	// Light dir matrix (on VU1 we can do 3 lights + Ambient for the same price as 1)
	VIFStaticDMA.AddVector(Vector4(LightDir1.x,LightDir2.x,LightDir3.x,0));				// 9
	VIFStaticDMA.AddVector(Vector4(LightDir1.y,LightDir2.y,LightDir3.y,0));				// 10
	VIFStaticDMA.AddVector(Vector4(LightDir1.z,LightDir2.z,LightDir3.z,0));				// 11

	// Light colour matrix (3 directional light colours)
	VIFStaticDMA.AddVector(Vector4(128  ,0,  0,  0));									// 12
	VIFStaticDMA.AddVector(Vector4(0  ,128,  0,  0));									// 13
	VIFStaticDMA.AddVector(Vector4(0  ,0,  128,  0));									// 14
	// Ambient light colour
	VIFStaticDMA.AddVector(Vector4(0, 0, 0,  0));									// 15


	// Add the GIFTag
	VIFStaticDMA.Add128(GS_GIFTAG_BATCH(	iVerts,						// NLOOP
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_LINE, 	// PRIM
													PRIM_IIP_FLAT, 
													PRIM_TME_OFF,
													PRIM_FGE_OFF, 
													PRIM_ABE_OFF, 
													PRIM_AA1_OFF, 
													PRIM_FST_STQ, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_3(	GIF_REG_ST,
	 													GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_XYZ2)));
	
	
	// X Axis
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4(-4, 0, 0, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 0, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4(-1, 0, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 4, 0, 0, 1));	// Vert
	
	// Y Axis
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0,-1, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 0,-4, 0, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 0, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0,-1, 0, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 0, 4, 0, 1));	// Vert
	
	// Z Axis
	VIFStaticDMA.AddVector(Vector4( 0, 0, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 0, 0,-4, 1));	// Vert
	VIFStaticDMA.AddVector(Vector4( 0, 1, 1, 0));   // STQ
	VIFStaticDMA.AddVector(Vector4( 0, 0,-1, 0));	// Normal
	VIFStaticDMA.AddVector(Vector4( 0, 0, 4, 1));	// Vert
	
	// Flush to make sure VU1 isn't doing anything.
	VIFStaticDMA.Add32(FLUSH);

	// Then run the microcode.
	VIFStaticDMA.Add32(MSCALL(0));

	// Return back to the dynamic chain (note that we aren't aligned on a QWORD boundary, but
	// it doesn't matter because the packet class will sort that out for us).
	VIFStaticDMA.DMARet();
	
	
}


Matrix4x4 GetCameraMatrix(void)
{
	// Position the camera at the origin
	static Vector4	CamPos(4, 2, 14, 1);
	
	// Set up a nice 3D camera
	static float yrot = 0.4f;
	static float xrot = -0.3f;
	
	// Reset the camer if requested
	if(pad[0].buttons & PAD_R2)
	{
		CamPos = Vector4(4, 2, 14, 1);
		yrot = 0.4f;
		xrot = -0.3f;
	}
	
	// Get any requested rotations
	yrot -= pad[0].axes[PAD_AXIS_RX] * Abs(pad[0].axes[PAD_AXIS_RX]) * 0.07f;
	xrot += pad[0].axes[PAD_AXIS_RY] * Abs(pad[0].axes[PAD_AXIS_RY]) * 0.07f;
	
	// Clamp xrot to prevent going upside down.
	if(xrot > ROTXMAX)
		xrot = ROTXMAX;
	
	if(xrot < -ROTXMAX)
		xrot = -ROTXMAX;
	
	// Sort out the rotations
	Matrix4x4 roty, rotx;
	roty.RotationY( yrot );
	rotx.RotationX( xrot );
	Matrix4x4 CamRot = rotx * roty;
	
	// Transpose rotation matrix to get camera matrix
	CamRot = Transpose(CamRot);

	// Get the direction vectors for the camera
	Vector4 vXAxis = Vector4(CamRot(0,0), CamRot(1,0), CamRot(2,0), 1);
	Vector4 vYAxis = Vector4(CamRot(0,1), CamRot(1,1), CamRot(2,1), 1);
	Vector4 vZAxis = Vector4(CamRot(0,2), CamRot(1,2), CamRot(2,2), 1);
	
	// Get any requested translations
	float strafe =   0.5f * pad[0].axes[PAD_AXIS_LX];
	float forwards = 0.5f * pad[0].axes[PAD_AXIS_LY];
	float updown =   0.5f * (pad[0].pressures[PAD_PL1] - pad[0].pressures[PAD_PL2]);

	// Work out the requested move for this frame
	Vector4 vReqMove;
	vReqMove =  vXAxis * strafe;
	vReqMove += vYAxis * updown;
	vReqMove += vZAxis * forwards;
	
	// And add it to the camera position
	CamPos += vReqMove;

	// Finally build the camera matrix
	Matrix4x4 matView;
	matView.Translation(-CamPos.x, -CamPos.y, -CamPos.z);
	matView = matView * CamRot;
	
	return matView;
}


int main()
{
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	// Initialise the memory
	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total
	VIFStaticDMA.Initialise(1536);	// 1536 * 4K Pages = 6MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA
	
	// Open the VU's 
	open("/dev/ps2vpu0", O_RDWR);
	open("/dev/ps2vpu1", O_RDWR);
	
	// Load the texture
	CTexture Tex1;
	Tex1.LoadBitmap("test.bmp");


	// Find out the length of the code in double words
	// We use double words since code is made up of many
	// pairs of two (1 upper, and 1 lower) VU instructions. 
	// Each instruction is 32 bits, so VU code must be a
	// multiple of 64 bits.
	uint32 iCodeLengthInDWORDs = (((uint32)&VU_vu1code_end - (uint32)&VU_vu1code_start) + 7) >> 3;
	// Copy the code to VU1_MICRO_MEM
	memcpy(VU1_MICRO_MEM, &VU_vu1code_start, iCodeLengthInDWORDs * 8);
	
	
	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);


	// Build a projection matrix
	float n = 2.0f;		// Near plane
	float f = 2000.0f;	// Far plane
	float FovYdiv2 = DegToRad(60.0f) * 0.5f; // 60 degree FOV
	float fAspect = 4.0f / 3.0f;			 // Aspect ratio
	float cotFOV = 1.0f / (Sin(FovYdiv2) / Cos(FovYdiv2));
	
	// We will be projecting to the 4096 wide drawing area
	// Look also at the VU projection code
	float w = cotFOV * (sps2UScreenGetWidth() / 4096.0f) / fAspect;
	float h = cotFOV * (sps2UScreenGetHeight() / 4096.0f);

	Matrix4x4 matProj
		(w, 0, 0, 0,
		 0, -h, 0, 0,
		 0, 0, (f+n) / (f-n), -1,
		 0, 0, (2*f*n) / (f-n), 0);
	
	// Get the address of static data that we can call to later.
	int iStaticCube = VIFStaticDMA.GetPointer();

	// Add the cube to the static buffer
	AddCube();
	
	// Get the address that we can call to later.
	int iAxesAddr = VIFStaticDMA.GetPointer();
	
	AddAxes();
		
	// Set up the DMA packet to clear the screen. We want to clear to blue.
	SPS2Manager.InitScreenClear(0x0, 0x05, 0x10);

	while(g_bLoop)
	{
		VIFDynamicDMA.Fire();
		
		pad_update(PAD_0);
		
		
		// Set up a world matrix that rotates the cube about it's own axis, and then again
		// around the origin.
		static float sfRotOrigin = 0.0f;
		static float sfRotLocalY = 0.0f;
		static float sfRotLocalZ = 0.0f;
		sfRotOrigin += pad[0].pressures[PAD_PLEFT] * 0.1f;
		sfRotOrigin -= pad[0].pressures[PAD_PRIGHT] * 0.1f;
		sfRotLocalY += 0.03f;
		sfRotLocalZ += 0.02f;

		Matrix4x4 matWorld, matTrans, matRotLocalY, matRotLocalZ, matRotOrigin;
		matTrans.Translation(0, 0, -10.0f);
		matRotOrigin.RotationY(sfRotOrigin);
		matRotLocalY.RotationY(sfRotLocalY);
		matRotLocalZ.RotationZ(sfRotLocalZ);

		matWorld = matRotLocalY * matRotLocalZ * matTrans * matRotOrigin;
			
		// Get the view matrix
		Matrix4x4 matView = GetCameraMatrix();
		
		// Get the combined view/projection matrix
		Matrix4x4 matViewProj = matView * matProj;	
		
		// You should be familiar with the idea of building 3D transforms from
		// the World * View * Projection matrices
		Matrix4x4 matWVP = matWorld * matViewProj;
		
		SPS2Manager.BeginScene();
		
		// Upload the current texture to the free texture buffer
		Tex1.Upload(TEXBUF480);
		Tex1.Select();
		
		// Upload the cube to render
		// We have to upload the matrix in the dynamic buffer
		// because it change every frame.
		VIFDynamicDMA.AddUnpack(V4_32, 0, 8);
		VIFDynamicDMA.AddMatrix(matWVP);
		VIFDynamicDMA.AddMatrix(matWorld);
		VIFDynamicDMA.DMACall(iStaticCube);

		// Upload the local axes for the cube
		// IDENTITY matrix is uploaded for the world matrix to get the axes colour right
		// Flush so that we wait for the previous cube to finish rendering
		// before we upload new data into that area of memory.
		VIFDynamicDMA.Add32(FLUSH);

		VIFDynamicDMA.AddUnpack(V4_32, 0, 8);
		VIFDynamicDMA.AddMatrix(matWVP);
		VIFDynamicDMA.AddMatrix(Matrix4x4::IDENTITY);
		VIFDynamicDMA.DMACall(iAxesAddr);


		// Upload the world axes to render
		// Flush so that we wait for the previous cube to finish rendering
		// before we upload new data into that area of memory.
		VIFDynamicDMA.Add32(FLUSH);

		VIFDynamicDMA.AddUnpack(V4_32, 0, 8);
		VIFDynamicDMA.AddMatrix(matViewProj);
		VIFDynamicDMA.AddMatrix(Matrix4x4::IDENTITY);
		VIFDynamicDMA.DMACall(iAxesAddr);
			
		SPS2Manager.EndScene();
		
		// Check for exit condition
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)) g_bLoop = false;
	}

	pad_cleanup(PAD_0);
	
	return 0;
	
}

